#!/usr/bin/env python

"""
A program which parses a MITAB file and outputs a number of separate files, each
representing a data table which can then be imported into a relational database
system.

Details of the employed tables are found in the accompanying 'init_mitab.sql'
file.

--------

Copyright (C) 2009, 2010, 2011 Ian Donaldson <ian.donaldson@biotek.uio.no>
Original author: Paul Boddie <paul.boddie@biotek.uio.no>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import os

try:
    set
except NameError:
    from sets import Set as set

# Regular expression patterns.

list_regexp = re.compile('(?P<field>(((?P<quote>").*?(?P=quote))|[^|])+)')

def int_or_none(x):
    try:
        return int(x)
    except ValueError:
        return None

# Classes.

class Parser:

    "A parser, maintaining references to the result files."

    filenames = [
        "mitab_interactions",
        "mitab_interactor_genes",
        "mitab_interactor_rogs",
        "mitab_canonical_interactor_rogs",
        "mitab_interaction_rigs",
        "mitab_canonical_interaction_rigs",
        "mitab_interactor_taxonomy",
        "mitab_alternatives",
        "mitab_aliases",
        "mitab_method_names",
        "mitab_interaction_type_names",
        "mitab_authors",
        "mitab_pubmed",
        "mitab_sources",
        "mitab_interaction_identifiers",
        "mitab_confidence"
        ]

    warnings_none, warnings_trivial, warnings_serious, warnings_all = range(0, 4)
    warning_labels = [None, "TRIVIAL", "SERIOUS"]
    integer_limit = 2**32/2

    def __init__(self, output_directory=None, warnings=None):

        "Initialise the result files."

        self.resultfiles = {}
        self.output_directory = output_directory or os.getcwd()
        self.warnings = warnings or self.warnings_none
        self.line_number = None

    def start(self):

        "Start parsing by opening the result files."

        if not os.path.exists(self.output_directory):
            os.mkdir(self.output_directory)

        for filename in self.filenames:
            self.resultfiles[filename] = open(os.path.join(self.output_directory, filename), "w")

    def close(self):

        "Close the result files."

        for filename in self.filenames:
            self.resultfiles[filename].close()

    def warning(self, text, level):

        "Output a warning using the given 'text' for the given 'level'."

        if self.warnings & level:
            print "%d: %s: %s" % (self.line_number, self.warning_labels[level], text)

    def parse(self, filename):

        "Parse the file with the given 'filename'."

        infile = open(filename)
        self.start()
        try:
            for self.line_number, line in enumerate(infile.xreadlines()):
                self.parse_line(line)
        finally:
            self.close()
            infile.close()

    def parse_line(self, line):

        "Parse the given 'line', writing results to the different result files."

        if line.startswith("#"):
            return

        fields = line.strip().split("\t")

        # Take only the recognised fields from the line.

        # For the original MITAB format...

        if len(fields) <= 24:

            (uidA, uidB,                                            # ROG identifiers (prefixed)
                altA, altB, aliasA, aliasB,                         # interactor details
                method, authors, pmids,
                taxA, taxB,
                interactionType,
                sourcedb, interactionIdentifiers,
                confidence,
                entrezGeneA, entrezGeneB,
                Atype, Btype,
                rigid,                                              # RIG identifier
                edgetype,
                numParticipants,
                irogidA, irogidB                                    # integer identifiers
                ) = fields[:24]

            # Convert various fields.

            entrezGeneA = self.convert_to_int("entrezGeneA", entrezGeneA, "-1")    # weak conversion with -1 as None
            entrezGeneB = self.convert_to_int("entrezGeneB", entrezGeneB, "-1")    # weak conversion with -1 as None
            entrezGenesA = entrezGeneA and [entrezGeneA] or []      # convert to a list
            entrezGenesB = entrezGeneB and [entrezGeneB] or []      # convert to a list

            irogidA = int(irogidA)
            irogidB = int(irogidB)

            # Some specific information is not provided.

            crogidA = None
            crogidB = None
            crigid = None

            mitab = 25

        # For the MITAB2.6 format...

        else:

            # The primary columns refer to "preferred" symbols for the
            # interactors.
            # NOTE: Currently such primary columns, along with various other
            # NOTE: columns, are ignored.

            # Various names match those above in order to gather information for
            # specific interactions and interactors.

            # Extract the standard fields first.

            (symA, symB,                                            # interactor symbols (or complex identifiers, prefixed)
                altA, altB, aliasA, aliasB,                         # interactor details
                method, authors, pmids,
                taxA, taxB,
                interactionType,
                sourcedb, interactionIdentifiers,
                confidence,
                # New in iRefIndex 7.0 MITAB2.6:
                expansion,
                biologicalRoleA, biologicalRoleB,
                experimentalRoleA, experimentalRoleB,
                # Preserved from the original format:
                Atype, Btype,
                # New in iRefIndex 7.0 MITAB2.6:
                xrefsA, xrefsB, xrefsInteraction,                   # not used by iRefIndex
                annotationsA, annotationsB, annotationsInteraction, # not used by iRefIndex
                taxHost, parametersInteraction,
                creationDate, updateDate,
                checksumA, checksumB, checksumInteraction,          # ROG and RIG identifiers, prefixed
                negative                                            # always "false" for iRefIndex
                ) = fields[:36]

            # Then extract any non-standard columns.

            if len(fields) >= 54:

                (origA, origB,                                          # the original identifiers
                    finalA, finalB,                                     # corrected/selected identifiers
                    mappingScoreA, mappingScoreB,
                    irogidA, irogidB,                                   # integer identifiers
                    irigid,
                    crogidA, crogidB,                                   # canonical ROG identifiers, unprefixed
                    crigid,                                             # canonical RIG identifier, unprefixed
                    icrogidA, icrogidB,                                 # integer identifiers
                    icrigid,
                    imexid,
                    # Preserved from the original format:
                    edgetype,
                    numParticipants,
                    ) = fields[36:54]

                # Permit the import of data so that the assignment of these
                # identifiers can be performed in a database.

                irogidA = int_or_none(irogidA)
                irogidB = int_or_none(irogidB)
                irigid = int_or_none(irigid)
                icrogidA = int_or_none(icrogidA)
                icrogidB = int_or_none(icrogidB)
                icrigid = int_or_none(icrigid)

            # Define missing non-standard columns as having no value.

            else:
                irogidA = None
                irogidB = None
                irigid = None
                crogidA = None
                crogidB = None
                crigid = None
                icrogidA = None
                icrogidB = None
                icrigid = None

                # NOTE: Assuming various values.

                edgetype = "X"
                numParticipants = 2

            # Convert various fields.

            uidA = checksumA
            uidB = checksumB
            rigid = self.convert_interaction(checksumInteraction)

            # Some specific information is not provided.

            entrezGenesA = []
            entrezGenesB = []

            mitab = 26

        # Common conversions.

        uidA_irefindex = self.convert_interactor(uidA)
        uidB_irefindex = self.convert_interactor(uidB)
        taxA_list = self.convert_taxids("taxA", taxA)
        taxB_list = self.convert_taxids("taxB", taxB)
        numParticipants = int(numParticipants)

        # Interactions.

        methods = self.convert_vocabulary_terms(method, "method")
        interaction_types = self.convert_vocabulary_terms(interactionType, "interaction type")

        # Write specific interaction details.
        # If a canonical RIG is known, this is specified.

        if not (uidA_irefindex and uidB_irefindex):
            self.warning("Incomplete interaction found!", self.warnings_serious)
        else:
            self.write_line(
                self.resultfiles["mitab_interactions"], (
                    uidA_irefindex, uidB_irefindex,
                    Atype, Btype, rigid, edgetype, numParticipants,
                    crogidA, crogidB, crigid)
                )

        # Include ROG information.

        if irogidA is not None:
            self.write_rog(uidA_irefindex, irogidA, 1)
        if irogidB is not None:
            self.write_rog(uidB_irefindex, irogidB, 1)

        # Include RIG information.

        if irigid is not None:
            self.write_rig(rigid, irigid, 1)

        # MITAB2.6-specific interaction details.

        if mitab == 26:

            # Include ROG information.

            if icrogidA is not None:
                self.write_rog(crogidA, icrogidA, 0)
            if icrogidB is not None:
                self.write_rog(crogidB, icrogidB, 0)

            # Include RIG information.

            if icrigid is not None:
                self.write_rig(crigid, icrigid, 0)

        # Include specific interactor information, if available.

        self.write_genes(uidA_irefindex, entrezGenesA)
        self.write_genes(uidB_irefindex, entrezGenesB)

        self.write_taxids(uidA_irefindex, taxA_list)
        self.write_taxids(uidB_irefindex, taxB_list)

        # Alternatives.

        mitab_alternatives = self.resultfiles["mitab_alternatives"]

        duplicates_found = self.write_alternatives(uidA_irefindex, altA)

        if uidA != uidB:
            duplicates_found = duplicates_found or self.write_alternatives(uidB_irefindex, altB)

        if duplicates_found:
            self.warning("Duplicate alternatives found: %s, %s, %s, %s" % (uidA, uidB, altA, altB), self.warnings_trivial)

        # Aliases.

        # Old-style interaction details.

        self.write_aliases(uidA_irefindex, aliasA)
        if uidA != uidB:
            self.write_aliases(uidB_irefindex, aliasB)

        # Methods.

        if not rigid and methods:
            self.warning("Method details provided for absent interaction.", self.warnings_serious)
        else:
            mitab_method_names = self.resultfiles["mitab_method_names"]

            for method_code, method_names in methods.items():
                for method_name in method_names:
                    self.write_line(mitab_method_names, (rigid, method_code, method_name))

        # Interaction types.

        if not rigid and interaction_types:
            self.warning("Interaction type details provided for absent interaction.", self.warnings_serious)
        else:
            mitab_interaction_type_names = self.resultfiles["mitab_interaction_type_names"]

            for interaction_type_code, interaction_type_names in interaction_types.items():
                for interaction_type_name in interaction_type_names:
                    self.write_line(mitab_interaction_type_names, (rigid, interaction_type_code, interaction_type_name))

        # Authors.

        author_list = self.convert_list(authors)

        if not rigid and author_list:
            self.warning("Author details provided for absent interaction.", self.warnings_serious)
        else:
            mitab_authors = self.resultfiles["mitab_authors"]
            author_set = set(author_list)

            for author in author_set:
                self.write_line(mitab_authors, (rigid, author))

            if len(author_list) != len(author_set):
                self.warning("Duplicate authors found for %s: %s" % (rigid, authors), self.warnings_trivial)

        # PubMed.

        pmid_list = self.convert_list(pmids)

        if not rigid and pmid_list:
            self.warning("PubMed details provided for absent interaction.", self.warnings_serious)
        else:
            mitab_pubmed = self.resultfiles["mitab_pubmed"]
            added_pubmed = set()

            for pmid in pmid_list:
                if pmid == "-":
                    continue

                prefix_value = pmid.split(":")

                if len(prefix_value) != 2:
                    self.warning("Inappropriate PubMed format: %s" % pmid, self.warnings_trivial)

                value = prefix_value[-1]

                if value in ("missing_pmid", "-1"):
                    continue

                try:
                    pmid = int(value)

                    if pmid >= self.integer_limit:
                        self.warning("Possible oversized PubMed identifier for %s: %s" % (rigid, pmid), self.warnings_serious)
                        continue

                    if pmid in added_pubmed:
                        self.warning("Duplicate PubMed identifier for %s: %s" % (rigid, pmid), self.warnings_trivial)
                        continue

                    self.write_line(mitab_pubmed, (rigid, pmid))
                    added_pubmed.add(pmid)

                except ValueError:
                    self.warning("Nonsense PubMed identifier for %s: %s" % (rigid, pmid), self.warnings_trivial)

        # Sources.

        db_list = self.convert_list(sourcedb)

        if not rigid and db_list:
            self.warning("Source information provided for absent interaction.", self.warnings_serious)
        else:
            mitab_sources = self.resultfiles["mitab_sources"]

            for db in db_list:
                sources = self.convert_vocabulary_terms(db, "database")

                assert len(sources.keys()) == 1
                source_code, source_names = sources.items()[0]

                if source_code is not None and source_code.endswith(":"):
                    source_code = source_code[:-1]
                    self.warning("Trailing data on source code for %s: %s" % (rigid, db), self.warnings_trivial)

                if not source_names:
                    self.warning("No source names accompanying code %s for %s: %s" % (source_code, rigid, db), self.warnings_trivial)
                else:
                    self.write_line(mitab_sources, (rigid, source_code, source_names[0]))

        # Interaction identifiers.

        identifier_list = self.convert_list(interactionIdentifiers)

        if not rigid and identifier_list:
            self.warning("Interaction identifiers provided for absent interaction.", self.warnings_serious)
        else:
            mitab_interaction_identifiers = self.resultfiles["mitab_interaction_identifiers"]

            for uid_orig in identifier_list:
                dbname, uid = self.convert_identifier(uid_orig, "interaction")
                if not uid:
                    self.warning("Interaction identifier for %s from database %s has no actual identifier: %s" % (rigid, dbname, uid_orig), self.warnings_trivial)
                    continue
                self.write_line(mitab_interaction_identifiers, (rigid, dbname, uid))

        # Confidence scores.

        confidence_list = self.convert_list(confidence)

        if not rigid and confidence_list:
            self.warning("Confidence scores provided for absent interaction.", self.warnings_serious)
        else:
            mitab_confidence = self.resultfiles["mitab_confidence"]

            for score in confidence_list:

                # Handle PSICQUIC anomoly.

                if score.startswith("PSICQUIC"):
                    continue

                if score == "-":
                    self.warning("Score entry has illegal format: %s" % score, self.warnings_serious)
                    continue

                type, value = score.split(":")
                if value == "-":
                    continue

                value = int(value) # strong conversion
                self.write_line(mitab_confidence, (rigid, type, value))

    def write_genes(self, uid, genes):

        """
        Write gene entries for the given 'uid' and 'genes'.
        """

        mitab_interactor_genes = self.resultfiles["mitab_interactor_genes"]

        for geneid in genes:
            self.write_line(mitab_interactor_genes, (uid, geneid))

    def write_taxids(self, uid, taxids):

        """
        Write taxonomy identifier entries for the given 'uid' and 'taxids'.
        """

        mitab_interactor_taxonomy = self.resultfiles["mitab_interactor_taxonomy"]

        for taxid in taxids:
            self.write_line(mitab_interactor_taxonomy, (uid, taxid))

    def write_rog(self, uid, rog, specific=1):

        """
        Write a redundant object group identifier entry for the given 'uid' and
        'rog' integer, according to whether it is 'specific' (a flag indicating
        a specific interactor if set to a true value).
        """

        if specific:
            mitab_interactor_rogs = self.resultfiles["mitab_interactor_rogs"]
        else:
            mitab_interactor_rogs = self.resultfiles["mitab_canonical_interactor_rogs"]

        self.write_line(mitab_interactor_rogs, (uid, rog))

    def write_rig(self, uid, rig, specific=1):

        """
        Write a redundant interaction group identifier entry for the given 'uid'
        and 'rig' integer, according to whether it is 'specific' (a flag
        indicating a specific interactor if set to a true value).
        """

        if specific:
            mitab_interaction_rigs = self.resultfiles["mitab_interaction_rigs"]
        else:
            mitab_interaction_rigs = self.resultfiles["mitab_canonical_interaction_rigs"]

        self.write_line(mitab_interaction_rigs, (uid, rig))

    def write_alternatives(self, uid, alts):

        """
        Write alternative symbol information for the given 'uid' and 'alts'
        column value.

        Return whether duplicates were found.
        """

        if not uid:
            self.warning("Alternatives provided for absent interactor.", self.warnings_serious)
            return

        mitab_alternatives = self.resultfiles["mitab_alternatives"]

        alt_list = self.convert_list(alts)
        alt_set = set(alt_list)
        for alt_orig in alt_set:
            dbname, alt = self.convert_identifier(alt_orig, "alternative")
            if not alt:
                self.warning("Alternative for %s from database %s has no actual identifier: %s" % (uid, dbname, alt_orig), self.warnings_trivial)
                continue
            if alt == uid:
                continue
            self.write_line(mitab_alternatives, (uid, dbname, alt))

        return len(alt_set) != len(alt_list)

    def write_aliases(self, uid, aliases):

        """
        Write records for the given 'uid' and 'aliases'.
        """

        if not uid:
            self.warning("Aliases provided for absent interactor.", self.warnings_serious)
            return

        mitab_aliases = self.resultfiles["mitab_aliases"]
        last_dbname = None
        added_aliases = set()

        for alias in self.convert_list(aliases):
            if alias == "irefindex:complex":
                continue

            if alias in added_aliases:
                self.warning("Alias duplicated: %s" % alias, self.warnings_trivial)
                continue

            dbname, dbalias = self.convert_identifier(alias, "alias")

            if not dbalias:
                self.warning("Alias without value: %s" % alias, self.warnings_trivial)
                continue

            if not dbname:
                if last_dbname is not None:
                    dbname = last_dbname
                    self.warning("Alias without source name (%s) adopted previous source name: %s" % (alias, last_dbname), self.warnings_trivial)
                else:
                    self.warning("Alias skipped due to missing source name: %s" % alias, self.warnings_serious)
            else:
                last_dbname = dbname

            self.write_line(mitab_aliases, (uid, dbname, dbalias))
            added_aliases.add(alias)

    def convert_rog(self, rog):

        "Convert the given 'rog' integer identifier field value."

        if rog.startswith("irefindex:"):
            return int(rog[10:])
        else:
            try:
                result = int(rog)
                return result
            except ValueError:
                self.warning("Redundant object group identifier is not an integer: %s" % rog, self.warnings_serious)
                return rog

    def convert_interaction(self, interaction):

        "Convert the given 'interaction' field value."

        if interaction.startswith("rigid:"):
            return interaction[6:]
        elif interaction.startswith("crigid:"):
            return interaction[7:]
        else:
            self.warning("Interaction prefix not specified: %s" % interaction, self.warnings_trivial)
            return interaction

    def convert_interactor(self, interactor):

        "Convert the given 'interactor' field value."

        if interactor.startswith("irefindex:"):
            return interactor[10:]
        elif interactor.startswith("rogid:"):
            return interactor[6:]
        elif interactor.startswith("crogid:"):
            return interactor[7:]
        else:
            self.warning("Interactor not defined by iRefIndex: %s" % interactor, self.warnings_serious)
            return interactor

    def convert_identifier(self, uid, identifier_type):

        """
        Convert the given 'uid' to a (database name, identifier) tuple, using
        the specified 'identifier_type' for error reporting, if necessary.
        """

        # Handle empty identifiers which probably should be absent in many cases
        # where identifiers appear in lists.

        if uid == "-":
            self.warning("%s identifier is given as empty: %s" % (identifier_type.capitalize(), uid), self.warnings_trivial)
            return "", None

        # Handle non-empty identifiers.

        uid_parts = uid.split(":")
        if len(uid_parts) < 2:
            self.warning("%s identifier has no prefix: %s" % (identifier_type.capitalize(), uid), self.warnings_serious)
            return "", uid
        elif len(uid_parts) == 2:
            return tuple(uid_parts)
        else:
            self.warning("%s identifier has complicated identifier: %s" % (identifier_type.capitalize(), uid), self.warnings_trivial)
            return uid_parts[0], ":".join(uid_parts[1:]).strip('"')

    def convert_list(self, value):

        "Convert the given list 'value' into a genuine list."

        if value in ("-1", "NA", "-"):
            if value in ("-1", "NA"):
                self.warning("Empty value given as %s" % value, self.warnings_trivial)
            return []
        else:
            values = []
            for m in list_regexp.finditer(value):
                v = m.group("field")
                if v:
                    v = v.strip('" ')
                    values.append(v)
            return values

    def convert_vocabulary_terms(self, terms, term_type):

        """
        Convert the vocabulary 'terms', returning a list of (code, names)
        tuples. The 'term_type' is used in error reporting, if necessary.
        """

        if terms == "NA":
            return {}

        if terms == "-":
            self.warning("For %s vocabulary term, empty value given as %s" % (term_type, terms), self.warnings_trivial)
            return {}

        results = {}

        for term in terms.split(")|"):
            code_names = term.split("(")

            if len(code_names) != 2:
                self.warning("Inappropriate %s vocabulary term format: %s" % (term_type, "(".join(code_names)), self.warnings_serious)
                code = code_names[0]
                names = ""
            else:
                code, names = code_names

            if names.endswith(")"):
                names = names[:-1]
            if not results.has_key(code):
                results[code] = []

            # Ignore "null" entries but retain "partially null" (non-empty name)
            # for reference (since this can apply to databases without an
            # assigned code).

            if code == "MI:0000":
                if names == "-":
                    self.warning("For %s vocabulary term, generic code %s has empty names section" % (term_type, code), self.warnings_trivial)
                    continue
                else:
                    self.warning("For %s vocabulary term, generic code %s has non-empty names section: %s" % (term_type, code, names), self.warnings_serious)

            # Complain about missing names.

            elif code != "MI:0000" and names == "-":
                self.warning("For %s vocabulary term, an empty names section is provided for code %s" % (term_type, code), self.warnings_serious)

            # Filter out unavailable names or empty names in a list.

            names = self.convert_list(names)

            if not names:
                self.warning("For %s vocabulary term, no names were found for code %s" % (term_type, code), self.warnings_serious)
                continue

            results[code] += names

        return results

    def convert_taxids(self, field, value):

        """
        For the given 'field', convert 'value' to a list of integers, discarding
        trailing taxonomy name details for each value given in a delimited
        field, and discarding values that represent null taxonomy identifiers.
        """

        taxids = []

        for taxid_plus_details in self.convert_list(value):

            # Remove "taxid:" and convert with NA as None.

            taxid_plus_details = value[6:].split("(")

            # No trailing details.

            if len(taxid_plus_details) == 1:
                taxid = taxid_plus_details[0]

            # Some trailing details.

            else: # len(taxid_plus_details) >= 2
                taxid = taxid_plus_details[0]

                details = "(".join(taxid_plus_details[1:])
                if not details.endswith(")"):
                    self.warning("Taxonomy details for %s not well-formed: %s" % (field, value), self.warnings_trivial)

            taxid = self.convert_to_int(field, taxid, warn=self.warnings_trivial)

            if taxid is not None:
                taxids.append(taxid)

        return taxids

    def convert_to_int(self, field, value, null_value=None, warn=0):

        """
        From the given 'field' type, convert 'value' to an integer, if possible,
        returning either the converted value, or None if a conversion was not
        appropriate or if the value signalled a "not applicable" or "absent"
        condition.

        The optional 'null_value' can be used to specify a value which shall be
        translated to None.

        The optional 'warn' flag, if set to a non-zero value, will issue the
        specified warning if an invalid value is processed.
        """

        # Test for common None value equivalents.

        if value in ("NA", "-"):
            return None

        # Where a specific null value is stated, return None for the given
        # value.

        if null_value is not None and value == null_value:
            return None

        # Convert to an integer. Some columns related to gene identifiers have
        # non-integer information inserted into them such as ROG identifiers;
        # such identifiers are discarded.

        try:
            return int(value)
        except ValueError:
            if warn:
                self.warning("Integer field %s does not provide an integer: %s" % (field, value), warn)
            return None

    def write_line(self, f, values):

        "Write to file 'f' the given 'values'."

        # Process the list of values, encoding each one.

        values = map(self.encode, values)

        # Write a tab-separated list of values plus a newline.

        f.write("\t".join(values) + "\n")

    def encode(self, value):

        "Encode a 'value' for result file output."

        if value is None:
            return "\\N"
        else:
            return str(value)

if __name__ == "__main__":
    import sys

    if len(sys.argv) < 2:
        print sys.argv[0], "<filename> [ <data_directory> ] [ --all | --serious | --trivial ]"
        sys.exit(1)

    warnings = (("--all" in sys.argv) and Parser.warnings_all or 0) | \
        (("--serious" in sys.argv) and Parser.warnings_serious or 0) | \
        (("--trivial" in sys.argv) and Parser.warnings_trivial or 0)

    # Use a new directory called data as the output directory by default.
    # NOTE: The directory name cannot begin with "--".

    if len(sys.argv) > 2 and not sys.argv[2].startswith("--"):
        directory = sys.argv[2]
    else:
        directory = "data"

    parser = Parser(directory, warnings=warnings)
    parser.parse(sys.argv[1])

# vim: tabstop=4 expandtab shiftwidth=4
